/******************************************************************************
 * FILE: bsp.c
 *
 * UC ARCHITECTURE:
 *    PowerPC
 *
 * UC FAMILY:
 *    SPC57
 *
 * DESCRIPTION:
 *    BSP package HW abstraction layer (HAL) API implementation.
 *
 * FILE HISTORY:
 *    V1.0    23.09.2017     RO
 *       > Initial version
 *    V2.0    25.01.2018     RO
 *       > File name rename and file header update
 *       > bsp_crt0_PreInit and PostInit functions added to simplify
 *         application crt0 hooks implementation
 *       > BSP_TIMER_SUPPORT configuration option added
 *       > BSP_CORE_ENABLE configuration option added
 *
 ******************************************************************************
 * Copyright (C) 2015-2018 HighTec EDV-Systeme GmbH.  All rights reserved.
 * This is proprietary software. Any use of the software requires a written
 * license agreement with HighTec EDV-Systeme GmbH. Please contact for
 * detailed license information: info@hightec-rt.com.
 *****************************************************************************/

/* Include BSP header file to get API declaration and types */
#include "bsp.h"


/* ============================================================================
 * BSP INTERRUPT SUPPORT
 * Implements its own Interrupt Vector Table in RAM that allows to
 * register application handlers in runtime trough defined API without
 * need to modify Low Level sources.
 * ==========================================================================*/

#if (BSP_ISR_TABLE)

/* Interrupt Vector Table
 * must be aligned on 4k address */
IsrHandler bsp_isr_table[1024] __attribute__ ((aligned (4096)));

#endif /* BSP_ISR_TABLE */


/* ============================================================================
 * BSP LOCAL FUNCTIONS DECLARATION
 * ==========================================================================*/

/* Supportive function for HW initialization */
static void bsp_cgm_Init(void);
static void bsp_swt_Init(void);
static void bsp_clk_Init(void);
static void bsp_pflash_Init(void);
static void bsp_pram_Init(void);

#if (BSP_ISR_TABLE)
static void bsp_isr_NonHandledInterrupt(void);
#endif

/* ============================================================================
 * uC crt0 Hooks
 * BSP default initialization in pre-init and post-init crt0 hooks.
 * They can serve as a template how such initialization can look like, if
 * the user wants to implement hooks sequence by his own.
 * ==========================================================================*/

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_crt0_PreInit()
 *    Mainly platform initialization (shared resources) executed
 *    by the RESET core. It runs before memories are initialized
 *    (ECC, clear and copy).
 * IMPORTANT:
 *    Routines called from this function CANNOT rely on initialized data
 *    and their use would probably finish in an exception due to ECC.
 * INPUTS:
 *    coreId: current execution core
 * --------------------------------------------------------------------------*/
void bsp_crt0_PreInit(uint32_t coreId)
{
	/* RESET Core to do the Platform initialization */
	if (coreId == UC_RESET_CORE)
	{
		bsp_uc_InitPlatform();
	}

	/* Core internal sub-systems, like cache etc.*/
	bsp_uc_InitCore();

}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_crt0_PostInit()
 *    Function finishes the startup initialization of shared resources,
 *    executed by the RESET core.
 *    Here the memory is fully functional with already initialized data.
 * NOTE:
 *    It is recommended to finish shared resource configuration before
 *    enabling inactive cores to avoid any timing issues related to
 *    fully asynchronous execution among the cores.
 * INPUTS:
 *   coreId: current execution core
 * --------------------------------------------------------------------------*/
void bsp_crt0_PostInit(uint32_t coreId)
{
	/* RESET Core to finish shared resource configuration */
	if (coreId == UC_RESET_CORE)
	{
		/* Initialize all vectors to bsp_isr_NonHandledInterrupt */
		#if (BSP_ISR_TABLE)
		bsp_isr_Init();
		#endif

		/* Enable PIT module for register access */
		#if (BSP_TIMER_SUPPORT)
		bsp_timer_EnableModule();
		#endif

		/* Enable all inactive cores */
		#if (BSP_CORE_ENABLE)
		bsp_uc_StartAllCores();
		#endif
	}

	/* Register interrupt table */
	#if (BSP_ISR_TABLE)
	bsp_uc_RegisterIsrTable(BSP_ISR_SW_MODE, (uint32_t *) bsp_isr_table);
	#endif

	/* Enable External (peripheral) interrupts */
	bsp_isr_EnableExternalInterrupts();
}


/* ============================================================================
 * BSP uC PLATFORM INITIALIZATION API
 * ==========================================================================*/

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_uc_InitPlatform()
 *    Main task is to initialize Clock system and other shared resources.
 *    This initialization shall run by one core only (e.g. the Reset Core).
 * CHARACTERISTICS:
 *    - to set system cgm to uC max frequency
 *    - to set cgm dividers to valid frequencies (max or below)
 *    - to enable most common peripherals cgms
 *    - to enable peripherals
 *    - to configure shared platform resources (PFlash, PRAM)
 *    - to enable Instruction Cache on execution Core
 * IMPORTANT:
 *    1. Global data might not initialized (including ECC) in the crt0
 *       pre-init phase. The code executed in this phase can rely only
 *       on Stack and automatic variables.
 *    2. Shall be called only ONCE in multi-core startup scenario
 * --------------------------------------------------------------------------*/
void bsp_uc_InitPlatform(void)
{
	/* Disable SWT(s) */
	bsp_swt_Init();

	/* Initialize CGM Clock system: Bridges and dividers */
	bsp_cgm_Init();

	/* Enable XTAL & PLL */
	bsp_clk_Init();

	/* Switch System Clock to PLL1 */
	uc_cgm_SwitchSystemClock(UC_SYSTEM_CLOCK_SEL);

	/* Enable all peripherals in DRUN mode */
	uc_me_EnableAllPeripheralsInDRUN();

	/* Configure PFLASH controller(s) */
	bsp_pflash_Init();

	/* Configure PRAM controller(s) */
	bsp_pram_Init();

	/* Enable SYSCLK0 Output */
	#if ((UC_MODULE_SYSCLK0) && (UC_SYSCLK0_ENABLE))
		bsp_clk_EnableSYSCLK0Output();
	#endif

	/* Enable SYSCLK1 Output */
	#if ((UC_MODULE_SYSCLK1) && (UC_SYSCLK1_ENABLE))
		bsp_clk_EnableSYSCLK1Output();
	#endif
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_uc_InitCore()
 *    uC core initialization.
 * --------------------------------------------------------------------------*/
void bsp_uc_InitCore(void)
{
	switch (uc_core_GetCurrentCore())
	{
	case 0:
		/* Instruction Cache */
		#if ((UC_CORE0_ICHACHE) && (UC_CORE0_ICACHE_ENABLE))
		uc_core_EnableICache();
		#endif
		/* Data Cache */
		#if ((UC_CORE0_DCHACHE) && (UC_CORE0_DCACHE_ENABLE))
		uc_core_EnableDCache();
		#endif
		break;
	case 1:
		/* Instruction Cache */
		#if ((UC_CORE1_ICHACHE) && (UC_CORE1_ICACHE_ENABLE))
		uc_core_EnableICache();
		#endif
		/* Data Cache */
		#if ((UC_CORE1_DCHACHE) && (UC_CORE1_DCACHE_ENABLE))
		uc_core_EnableDCache();
		#endif
		break;

	case 2:
		/* Instruction Cache */
		#if ((UC_CORE2_ICHACHE) && (UC_CORE2_ICACHE_ENABLE))
		uc_core_EnableICache();
		#endif
		/* Data Cache */
		#if ((UC_CORE2_DCHACHE) && (UC_CORE2_DCACHE_ENABLE))
		uc_core_EnableDCache();
		#endif
		break;
	}
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_uc_StartCore()
 *    Starts requested Core with default setting of the reset_vector
 *    to _crt0_reset() entry point. If user wants a different reset vector,
 *    he needs to call LLD function directly.
 * INPUT:
 *    core_nb : core to start
 * --------------------------------------------------------------------------*/
extern void uc_StartCore(uint32_t core_nb);

void bsp_uc_StartCore(uint32_t core_nb)
{
	/* call uC specific version of Start Core */
	uc_StartCore(core_nb);

}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_uc_StartAllCores()
 *    Starts all inactive cores on multi-core uC derivative.
 *    Active ones are left untouched.
 *    All inactive cores will start at _crt0_reset entry point.
 * NOTE:
 *    All cores share the same crt0 startup routine.
 * --------------------------------------------------------------------------*/
void bsp_uc_StartAllCores(void)
{
	uint32_t i;
	for (i=0; i < UC_CORE_MAX; i++)
		bsp_uc_StartCore(i);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_uc_GetCurrentCoreId()
 *    Return current CoreId on which this call executes.
 * --------------------------------------------------------------------------*/
uint32_t bsp_uc_GetCurrentCoreId(void)
{
	return uc_core_GetCurrentCore();
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_uc_RegisterIsrTable()
 *    Core interrupt BSP ISR table registration.
 *    In the end Core's current priority is set to minimum to unblock all
 *    interrupt requests with higher priority than minimum. Default value
 *    of current priority after the reset is MAXIMUM that would block
 *    all pending interrupt requests.
 * INPUT
 *    isr_mode  : 0/1 SW/HW Vector mode
 *    isr_table : address of the Interrupt vector table
 * --------------------------------------------------------------------------*/
void bsp_uc_RegisterIsrTable(uint32_t isr_mode, uint32_t* isr_table)
{
	/* Read current core where to register ISR table */
	uint32_t currentCore = uc_core_GetCurrentCore();

	/* Set Interrupt Vector mode */
	uc_intc_SetBCR(currentCore, isr_mode);

	/* Register BSP Vector table in INTC module */
	uc_intc_SetVTBA(currentCore, isr_table);

	/* Set Core's current priority to minimum */
	uc_intc_SetCurrentPriority(currentCore, 0);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_clk_EnableSYSCLK0Output()
 *    Enables selected cgm to go on the dedicated pin that it
 *    can be measured externally, which might help to detect mismatch
 *    in cgm configuration.
 *    Corresponding pin is configured 'Strong' to support higher frequencies.
 * INPUT:
 *    none
 * --------------------------------------------------------------------------*/
#if (UC_MODULE_SYSCLK0)
void bsp_clk_EnableSYSCLK0Output(void)
{
#if defined(UC_SYSCLK0_AUX)
	/* SYSCLK0 on AUX bridge */
	uc_cgm_SelectAuxClock(UC_SYSCLK0_AUX, UC_SYSCLK0_SEL);
	uc_cgm_SetAuxClockDiv(UC_SYSCLK0_AUX, UC_SYSCLK0_DC, UC_SYSCLK0_DIV);
#else
	/* SYSCLKK0 on SC bridge */
	uc_cgm_SetScClockDiv(UC_SYSCLK0_DC, UC_SYSCLK0_DIV);
#endif

	/* Pin enable */
	uc_siul_EnableOutput(UC_SYSCLK0_PIN, UC_SIUL_OERC_STRONG, UC_SIUL_TYPE_OUTPUT_PUSH_PULL, 1);
}
#endif

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_clk_EnableSYSCLK1Output()
 *    Enables selected cgm to go on the dedicated pin that it
 *    can be measured externally, which might help to detect mismatch
 *    in cgm configuration.
 *    Corresponding pin is configured 'Strong' to support higher frequencies.
 * INPUT:
 *    none
 * --------------------------------------------------------------------------*/
#if (UC_MODULE_SYSCLK1)
void bsp_clk_EnableSYSCLK1Output(void)
{
#if defined(UC_SYSCLK1_AUX)
	/* SYSCLK1 on AUX bridge */
	uc_cgm_SelectAuxClock(UC_SYSCLK1_AUX, UC_SYSCLK1_SEL);
	uc_cgm_SetAuxClockDiv(UC_SYSCLK1_AUX, UC_SYSCLK1_DC, UC_SYSCLK1_DIV);
#else
	/* SYSCLK1 on SC bridge */
	uc_cgm_SetScClockDiv(UC_SYSCLK1_DC, UC_SYSCLK1_DIV);
#endif

	/* Pin enable */
	uc_siul_EnableOutput(UC_SYSCLK1_PIN, UC_SIUL_OERC_STRONG, UC_SIUL_TYPE_OUTPUT_PUSH_PULL, 1);
}
#endif


/* ============================================================================
 * BSP PIN API
 * ==========================================================================*/

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_pin_EnableInput()
 *    Enables Input mode on the pin with following characteristics
 *    - no pull-up
 *    - no pull-down.
 *    If pull-ups/downs are needed, user has to call LLD API directly
 *    from the application code
 * INPUT:
 *    pin_nb : Requested Pin
 * -------------------------------------------------------------------------- */
void bsp_pin_EnableInput(uint32_t pin_nb)
{
	uc_siul_EnableInput(pin_nb, UC_SIUL_PULL_DISABLE, UC_SIUL_PULL_DISABLE);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_pin_EnableOutput()
 *    Enables Output pin mode on the pin with following characteristics
 *    - edge rate = medium strength that shall be sufficient for many cases
 *    - push-pull output
 *    If other output pin characteristics are needed, user has to call
 *    LLD API directly from the application code
 * INPUT:
 *    pin_nb : Requested Pin
 *    source : Module that drives the pin (see Pin excel sheet attached to RM)
 * --------------------------------------------------------------------------*/
void bsp_pin_EnableOutput(uint32_t pin_nb, uint32_t source)
{
	uc_siul_EnableOutput(pin_nb, UC_SIUL_OERC_MEDIUM, UC_SIUL_TYPE_OUTPUT_PUSH_PULL, source);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_pin_EnableBidir()
 *    Enables Output pin mode on the pin with weak edge rate control and
 *    push-pull type.
 *    It enables Input pin mode on the same pin that user can read
 *    output signal back.
 *    If other pin characteristics are needed, user has to call
 *    LLD API directly from the application code.
 * INPUT:
 *    pin_nb : Requested Pin
 *    source : Module that drives the pin (see Pin excel sheet attached to RM)
 * NOTE:
 *    The enable sequence, first output followed by input enable, is important.
 * --------------------------------------------------------------------------*/
void bsp_pin_EnableBidir(uint32_t pin_nb, uint32_t source)
{
	uc_siul_EnableOutput(pin_nb, UC_SIUL_OERC_MEDIUM, UC_SIUL_TYPE_OUTPUT_PUSH_PULL, source);
	uc_siul_EnableInput(pin_nb, UC_SIUL_PULL_DISABLE, UC_SIUL_PULL_DISABLE);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_pin_SetOutput()
 *    Set the pin output to requested level
 * INPUT:
 *    pin_nb : Requested Pin
 *    value  : 0/1 = Low/High
 * --------------------------------------------------------------------------*/
void bsp_pin_SetOutput(uint32_t pin_nb, uint8_t value)
{
	uc_siul_SetGPIO(pin_nb, value);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_pin_GetInput()
 *    Get the pin input
 * INPUT:
 *    pin_nb : Requested Pin
 * RETURN:
 *    value  : 0/1 = Low/High
 * --------------------------------------------------------------------------*/
uint8_t bsp_pin_GetInput(uint32_t pin_nb)
{
	return uc_siul_GetGPIO(pin_nb);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_pin_Toggle()
 *    Alternates value on the pin, by "read - invert - write back"
 * INPUT:
 *    pin_nb : Requested Pin
 * --------------------------------------------------------------------------*/
void bsp_pin_Toggle(uint32_t pin_nb)
{
	uc_siul_SetGPIO(pin_nb, ~uc_siul_GetGPIO(pin_nb));
}


/* ============================================================================
 * BSP INTERRUPT API
 * ==========================================================================*/

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_isr_Init()
 *    Initialization of shared Interrupt table implemented in BSP.
 * INPUT:
 *    none
 * IMPORTANT:
 *    It must be done only once by one core!
 * --------------------------------------------------------------------------*/
void bsp_isr_Init(void)
{
#if (BSP_ISR_TABLE)

	uint32_t i;

	/* Initialize Cores' shared interrupt table, done by RESET core */
	for (i=0; i < 1024; i++)
		bsp_isr_table[i] = bsp_isr_NonHandledInterrupt;
#endif
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_isr_RegisterHandler()
 *    Registration of the ISR handler for a given vector (peripheral source).
 *    Two step operation
 *    1. to update a corresponding field in BSP ISR vector table
 *    2. update INTC module for a given vector number with the info
 *       a) which core will serve it (now it is current execution core)
 *       b) vector priority
 * INPUT:
 *    vector_nb : interrupt source (see RM for the mapping)
 *    priority  : interrupt source priority
 *    handler   : address of the interrupt handler function
 * --------------------------------------------------------------------------*/
void bsp_isr_RegisterHandler(uint32_t vector_nb, uint32_t priority, IsrHandler handler)
{
    /* register handler in Vector Table */
	#if (BSP_ISR_TABLE)
	bsp_isr_table[vector_nb] = handler;
	#else
	(void) handler;
	#endif


	/* configure INTC module PSR register */
	uc_intc_SetPSR(vector_nb, uc_core_GetCurrentCore(), priority);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_isr_EnableExternalInterrupts()
 *    Enables external interrupt source in the current Core's MSR register
 * --------------------------------------------------------------------------*/
void bsp_isr_EnableExternalInterrupts(void)
{
	asm(" wrteei 1 ");
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_isr_DisableExternalInterrupts()
 *    Disables external interrupt source in the current Core's MSR register
 * --------------------------------------------------------------------------*/
void bsp_isr_DisableExternalInterrupts(void)
{
	asm(" wrteei 0 ");
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_isr_ExitHWInterrupt()
 *    Sequence to finish interrupt in HW Vector mode before ISR clean up and
 *    call rfi instruction
 * --------------------------------------------------------------------------*/
void bsp_isr_ExitHWInterrupt(void)
{
	asm(" wrteei 0 ");
	uc_intc_EndOfInterrupt(uc_core_GetCurrentCore());
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_isr_NonHandledInterrupt()
 *    Dummy ISR handler providing infinitive loop when called, which
 *    should not occur under normal operation
 * --------------------------------------------------------------------------*/
#if (BSP_ISR_TABLE)
void bsp_isr_NonHandledInterrupt(void)
{
	while(1);
}
#endif


/* ============================================================================
 * BSP TIMER
 * ==========================================================================*/

#if (BSP_TIMER_SUPPORT)

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_timer_EnableModule()
 *    Enables BSP supported PIT module (one module with multiple channels).
 * --------------------------------------------------------------------------*/
void bsp_timer_EnableModule(void)
{
	uc_pit_Enable();
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_timer_ReloadChannel()
 *    Setting reload value for PIT channel.
 * NOTE:
 *    Runtime reload without stopping the channel. New reload value is
 *    applied by the PIT module during timer under-run (count down to 0).
 *    In case new value shall be applied immediately, timer must be stopped,
 *    reloaded and started again.
 * INPUT:
 *    channel  : requested PIT channel
 *    period   : time period in PER_CLOCK ticks (see Arch. BSP header file)
 * --------------------------------------------------------------------------*/
void bsp_timer_ReloadChannel(UC_PIT_CHANNEL_e channel, uint32_t period)
{
	uc_pit_ReloadChannel(channel, period);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_timer_EnableChannelIsr()
 *    Enable of PIT interrupt from a given channel.
 * NOTE:
 *    It is recommended to stop the timer, clear ISR flag, enable ISR and
 *    start the timer.
 * INPUT:
 *    channel  : requested PIT channel
 *    priority : interrupt priority
 *    handler  : interrupt handler
 * --------------------------------------------------------------------------*/
void bsp_timer_EnableChannelIsr(UC_PIT_CHANNEL_e channel, uint32_t priority, IsrHandler handler)
{
	uc_pit_EnableChannelIsr(channel);
	bsp_isr_RegisterHandler(UC_PIT_CH0_ISR_VECTOR + channel, priority, handler);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_timer_StartChannel()
 *    Starts timer
 * INPUT:
 *    channel : requested PIT channel
 * --------------------------------------------------------------------------*/
void bsp_timer_StartChannel(UC_PIT_CHANNEL_e channel)
{
	uc_pit_StartChannel(channel);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_timer_StopChannel(channel)
 *    Stops timer. Before stopping it reads the current counter value
 *    to return it if one needs it. Value needs to be read when the PIT timer
 *    is active.
 * INPUT:
 *    channel : requested PIT channel
 * OUTPUT:
 *    uint32 : value before stopping the timer
 * --------------------------------------------------------------------------*/
uint32_t bsp_timer_StopChannel(UC_PIT_CHANNEL_e channel)
{
	uint32_t value;
	value = uc_pit_GetChannelCurrentValue(channel);
	uc_pit_StopChannel(channel);
	return value;
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_timer_ClearChannelIsrFlag()
 *    Clears timer interrupt flag.
 * INPUT:
 *    channel : requested PIT channel
 * --------------------------------------------------------------------------*/
void bsp_timer_ClearChannelIsrFlag(UC_PIT_CHANNEL_e channel)
{
	uc_pit_ClearChannelIsrFlag(channel);
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_timer_IsChannelIsrFlagActive()
 *    Reads back the interrupt flag status
 * INPUT:
 *    channel : requested PIT channel
 * OUTPUT:
 *    uint32 : 0 = not active (still the interval has not elapsed)
 *             1 = active (period is over)
 * --------------------------------------------------------------------------*/
uint32_t bsp_timer_IsChannelIsrFlagActive(UC_PIT_CHANNEL_e channel)
{
	return uc_pit_IsChannelIsrFlag(channel);
}

#endif /* BSP_TIMER_SUPPORT */


/* ============================================================================
 * LOCAL SUPPORTIVE FUNCTIONS
 * ==========================================================================*/

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_swt_Init()
 *    Disables SWT Watchdog(S)
 * --------------------------------------------------------------------------*/
static void bsp_swt_Init(void)
{
	#if (UC_SWT0_DISABLE)
		uc_swt_Disable(UC_SWT0_BASE);
	#endif

	/* Disable SWT_1 */
	#if (UC_SWT1_DISABLE)
		uc_swt_Disable(UC_SWT1_BASE);
	#endif

	/* Disable SWT_2 */
	#if (UC_SWT2_DISABLE)
		uc_swt_Disable(UC_SWT2_BASE);
	#endif

	/* Disable SWT_3 */
	#if (UC_SWT3_DISABLE)
		uc_swt_Disable(UC_SWT3_BASE);
	#endif
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_clk_Init()
 *    It does initialization of basic cgm elements
 *    1. External oscillator (XOSC)
 *    2. PLL0 (if module exists)
 *    3. PLL1 (if module exists)
 * --------------------------------------------------------------------------*/
static void bsp_clk_Init(void)
{
	uc_cgm_EnableXOSC();

	/* Enable PLLs */
	#if (UC_MODULE_PLL0)
		uc_cgm_EnablePLL(0, UC_PLL0_PREDIV, UC_PLL0_MFD, UC_PLL0_RFDPHI, UC_PLL0_RFDPHI1);
	#endif
	#if (UC_MODULE_PLL1)
		uc_cgm_EnablePLL(1, 0, UC_PLL1_MFD, UC_PLL1_RFDPHI, 0);
	#endif
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_pflash_Init()
 *    It does initialization of implemented PFLASH controller
 * --------------------------------------------------------------------------*/
static void bsp_pflash_Init(void)
{
#if (UC_MODULE_PFLASH0)
	uc_pflash_SetPFCR1(UC_PFLASH0_BASE, UC_PFLASH0_PFCR1_APC, UC_PFLASH0_PFCR1_WS, UC_PFLASH0_PFCR1_PFLIM);
	uc_pflash_SetPFCR2(UC_PFLASH0_BASE, UC_PFLASH0_PFCR2_PFLIM);
	uc_pflash_SetPFCR3(UC_PFLASH0_BASE, UC_PFLASH0_PFCR3_P0WCFG, UC_PFLASH0_PFCR3_P1WCFG, UC_PFLASH0_PFCR3_ARB);
#endif

#if (UC_MODULE_PFLASH1)
	uc_pflash_SetPFCR1(UC_PFLASH1_BASE, UC_PFLASH1_PFCR1_APC, UC_PFLASH1_PFCR1_WS, UC_PFLASH1_PFCR1_PFLIM);
	uc_pflash_SetPFCR2(UC_PFLASH1_BASE, UC_PFLASH1_PFCR2_PFLIM);
	uc_pflash_SetPFCR3(UC_PFLASH1_BASE, UC_PFLASH1_PFCR3_P0WCFG, UC_PFLASH1_PFCR3_P1WCFG, UC_PFLASH1_PFCR3_ARB);
#endif
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_pram_Init()
 *    It does initialization of implemented PRAM controller
 * --------------------------------------------------------------------------*/
static void bsp_pram_Init(void)
{
#if (UC_MODULE_PRAM0)
	uc_pram_SetPRCR1(UC_PRAM0_BASE, UC_PRAM0_PFCR1_PRI, UC_PRAM0_PFCR1_P0RB, UC_PRAM0_PFCR1_P1RB, UC_PRAM0_PFCR1_FTDIS);
#endif

#if (UC_MODULE_PRAM1)
	uc_pram_SetPRCR1(UC_PRAM1_BASE, UC_PRAM1_PFCR1_PRI, UC_PRAM1_PFCR1_P0RB, UC_PRAM1_PFCR1_P1RB, UC_PRAM1_PFCR1_FTDIS);
#endif

#if (UC_MODULE_PRAM2)
	uc_pram_SetPRCR1(UC_PRAM2_BASE, UC_PRAM2_PFCR1_PRI, UC_PRAM2_PFCR1_P0RB, UC_PRAM2_PFCR1_P1RB, UC_PRAM2_PFCR1_FTDIS);
#endif

#if (UC_MODULE_PRAM3)
	uc_pram_SetPRCR1(UC_PRAM3_BASE, UC_PRAM3_PFCR1_PRI, UC_PRAM3_PFCR1_P0RB, UC_PRAM3_PFCR1_P1RB, UC_PRAM3_PFCR1_FTDIS);
#endif
}

/* ----------------------------------------------------------------------------
 * FUNCTION: bsp_cgm_Init()
 *    It does initialization of configured
 *    - SC DC Dividers
 *    - AUX Bridges with their Dividers
 * --------------------------------------------------------------------------*/
static void bsp_cgm_Init(void)
{
/* SC DC */
#if (UC_CGM_SC_DC0_INIT)
	uc_cgm_SetScClockDiv(0, UC_CGM_SC_DC0_DIV);
#endif
#if (UC_CGM_SC_DC1_INIT)
	uc_cgm_SetScClockDiv(1, UC_CGM_SC_DC1_DIV);
#endif
#if (UC_CGM_SC_DC2_INIT)
	uc_cgm_SetScClockDiv(2, UC_CGM_SC_DC2_DIV);
#endif
#if (UC_CGM_SC_DC3_INIT)
	uc_cgm_SetScClockDiv(3, UC_CGM_SC_DC3_DIV);
#endif
#if (UC_CGM_SC_DC4_INIT)
	uc_cgm_SetScClockDiv(4, UC_CGM_SC_DC4_DIV);
#endif
#if (UC_CGM_SC_DC5_INIT)
	uc_cgm_SetScClockDiv(5, UC_CGM_SC_DC5_DIV);
#endif

/* CGM AUX0 */
#if (UC_CGM_AC0_INIT)
	uc_cgm_SelectAuxClock(0, UC_CGM_AC0_CLOCK_INPUT);
#endif
#if (UC_CGM_AC0_DC0_INIT)
	uc_cgm_SetAuxClockDiv(0, 0, UC_CGM_AC0_DC0_DIV);
#endif
#if (UC_CGM_AC0_DC1_INIT)
	uc_cgm_SetAuxClockDiv(0, 1, UC_CGM_AC0_DC1_DIV);
#endif
#if (UC_CGM_AC0_DC2_INIT)
	uc_cgm_SetAuxClockDiv(0, 2, UC_CGM_AC0_DC2_DIV);
#endif
#if (UC_CGM_AC0_DC3_INIT)
	uc_cgm_SetAuxClockDiv(0, 3, UC_CGM_AC0_DC3_DIV);
#endif
#if (UC_CGM_AC0_DC4_INIT)
	uc_cgm_SetAuxClockDiv(0, 4, UC_CGM_AC0_DC4_DIV);
#endif
#if (UC_CGM_AC0_DC5_INIT)
	uc_cgm_SetAuxClockDiv(0, 5, UC_CGM_AC0_DC5_DIV);
#endif

/* CGM AUX1 */
#if (UC_CGM_AC1_INIT)
	uc_cgm_SelectAuxClock(1, UC_CGM_AC1_CLOCK_INPUT);
#endif
#if (UC_CGM_AC1_DC0_INIT)
	uc_cgm_SetAuxClockDiv(1, 0, UC_CGM_AC1_DC0_DIV);
#endif
#if (UC_CGM_AC1_DC1_INIT)
	uc_cgm_SetAuxClockDiv(1, 1, UC_CGM_AC1_DC1_DIV);
#endif
#if (UC_CGM_AC1_DC2_INIT)
	uc_cgm_SetAuxClockDiv(1, 2, UC_CGM_AC1_DC2_DIV);
#endif
#if (UC_CGM_AC1_DC3_INIT)
	uc_cgm_SetAuxClockDiv(1, 3, UC_CGM_AC1_DC3_DIV);
#endif
#if (UC_CGM_AC1_DC4_INIT)
	uc_cgm_SetAuxClockDiv(1, 4, UC_CGM_AC1_DC4_DIV);
#endif
#if (UC_CGM_AC1_DC5_INIT)
	uc_cgm_SetAuxClockDiv(1, 5, UC_CGM_AC1_DC5_DIV);
#endif

/* CGM AUX2 */
#if (UC_CGM_AC2_INIT)
	uc_cgm_SelectAuxClock(2, UC_CGM_AC2_CLOCK_INPUT);
#endif
#if (UC_CGM_AC2_DC0_INIT)
	uc_cgm_SetAuxClockDiv(2, 0, UC_CGM_AC2_DC0_DIV);
#endif
#if (UC_CGM_AC2_DC1_INIT)
	uc_cgm_SetAuxClockDiv(2, 1, UC_CGM_AC2_DC1_DIV);
#endif
#if (UC_CGM_AC2_DC2_INIT)
	uc_cgm_SetAuxClockDiv(2, 2, UC_CGM_AC2_DC2_DIV);
#endif
#if (UC_CGM_AC2_DC3_INIT)
	uc_cgm_SetAuxClockDiv(2, 3, UC_CGM_AC2_DC3_DIV);
#endif
#if (UC_CGM_AC2_DC4_INIT)
	uc_cgm_SetAuxClockDiv(2, 4, UC_CGM_AC2_DC4_DIV);
#endif
#if (UC_CGM_AC2_DC5_INIT)
	uc_cgm_SetAuxClockDiv(2, 5, UC_CGM_AC2_DC5_DIV);
#endif

/* CGM AUX3 */
#if (UC_CGM_AC3_INIT)
	uc_cgm_SelectAuxClock(3, UC_CGM_AC3_CLOCK_INPUT);
#endif
#if (UC_CGM_AC3_DC0_INIT)
	uc_cgm_SetAuxClockDiv(3, 0, UC_CGM_AC3_DC0_DIV);
#endif
#if (UC_CGM_AC3_DC1_INIT)
	uc_cgm_SetAuxClockDiv(3, 1, UC_CGM_AC3_DC1_DIV);
#endif
#if (UC_CGM_AC3_DC2_INIT)
	uc_cgm_SetAuxClockDiv(3, 2, UC_CGM_AC3_DC2_DIV);
#endif
#if (UC_CGM_AC3_DC3_INIT)
	uc_cgm_SetAuxClockDiv(3, 3, UC_CGM_AC3_DC3_DIV);
#endif
#if (UC_CGM_AC3_DC4_INIT)
	uc_cgm_SetAuxClockDiv(3, 4, UC_CGM_AC3_DC4_DIV);
#endif
#if (UC_CGM_AC3_DC5_INIT)
	uc_cgm_SetAuxClockDiv(3, 5, UC_CGM_AC3_DC5_DIV);
#endif

/* CGM AUX4 */
#if (UC_CGM_AC4_INIT)
	uc_cgm_SelectAuxClock(4, UC_CGM_AC4_CLOCK_INPUT);
#endif
#if (UC_CGM_AC4_DC0_INIT)
	uc_cgm_SetAuxClockDiv(4, 0, UC_CGM_AC4_DC0_DIV);
#endif
#if (UC_CGM_AC4_DC1_INIT)
	uc_cgm_SetAuxClockDiv(4, 1, UC_CGM_AC4_DC1_DIV);
#endif
#if (UC_CGM_AC4_DC2_INIT)
	uc_cgm_SetAuxClockDiv(4, 2, UC_CGM_AC4_DC2_DIV);
#endif
#if (UC_CGM_AC4_DC3_INIT)
	uc_cgm_SetAuxClockDiv(4, 3, UC_CGM_AC4_DC3_DIV);
#endif
#if (UC_CGM_AC4_DC4_INIT)
	uc_cgm_SetAuxClockDiv(4, 4, UC_CGM_AC4_DC4_DIV);
#endif
#if (UC_CGM_AC4_DC5_INIT)
	uc_cgm_SetAuxClockDiv(4, 5, UC_CGM_AC4_DC5_DIV);
#endif

/* CGM AUX5 */
#if (UC_CGM_AC5_INIT)
	uc_cgm_SelectAuxClock(5, UC_CGM_AC5_CLOCK_INPUT);
#endif
#if (UC_CGM_AC5_DC0_INIT)
	uc_cgm_SetAuxClockDiv(5, 0, UC_CGM_AC5_DC0_DIV);
#endif
#if (UC_CGM_AC5_DC1_INIT)
	uc_cgm_SetAuxClockDiv(5, 1, UC_CGM_AC5_DC1_DIV);
#endif
#if (UC_CGM_AC5_DC2_INIT)
	uc_cgm_SetAuxClockDiv(5, 2, UC_CGM_AC5_DC2_DIV);
#endif
#if (UC_CGM_AC5_DC3_INIT)
	uc_cgm_SetAuxClockDiv(5, 3, UC_CGM_AC5_DC3_DIV);
#endif
#if (UC_CGM_AC5_DC4_INIT)
	uc_cgm_SetAuxClockDiv(5, 4, UC_CGM_AC5_DC4_DIV);
#endif
#if (UC_CGM_AC5_DC5_INIT)
	uc_cgm_SetAuxClockDiv(5, 5, UC_CGM_AC5_DC5_DIV);
#endif

/* CGM AUX6 */
#if (UC_CGM_AC6_INIT)
	uc_cgm_SelectAuxClock(6, UC_CGM_AC6_CLOCK_INPUT);
#endif
#if (UC_CGM_AC6_DC0_INIT)
	uc_cgm_SetAuxClockDiv(6, 0, UC_CGM_AC6_DC0_DIV);
#endif
#if (UC_CGM_AC6_DC1_INIT)
	uc_cgm_SetAuxClockDiv(6, 1, UC_CGM_AC6_DC1_DIV);
#endif
#if (UC_CGM_AC6_DC2_INIT)
	uc_cgm_SetAuxClockDiv(6, 2, UC_CGM_AC6_DC2_DIV);
#endif
#if (UC_CGM_AC6_DC3_INIT)
	uc_cgm_SetAuxClockDiv(6, 3, UC_CGM_AC6_DC3_DIV);
#endif
#if (UC_CGM_AC6_DC4_INIT)
	uc_cgm_SetAuxClockDiv(6, 4, UC_CGM_AC6_DC4_DIV);
#endif
#if (UC_CGM_AC6_DC5_INIT)
	uc_cgm_SetAuxClockDiv(6, 5, UC_CGM_AC6_DC5_DIV);
#endif

/* CGM AUX7 */
#if (UC_CGM_AC7_INIT)
	uc_cgm_SelectAuxClock(7, UC_CGM_AC7_CLOCK_INPUT);
#endif
#if (UC_CGM_AC7_DC0_INIT)
	uc_cgm_SetAuxClockDiv(7, 0, UC_CGM_AC7_DC0_DIV);
#endif
#if (UC_CGM_AC7_DC1_INIT)
	uc_cgm_SetAuxClockDiv(7, 1, UC_CGM_AC7_DC1_DIV);
#endif
#if (UC_CGM_AC7_DC2_INIT)
	uc_cgm_SetAuxClockDiv(7, 2, UC_CGM_AC7_DC2_DIV);
#endif
#if (UC_CGM_AC7_DC3_INIT)
	uc_cgm_SetAuxClockDiv(7, 3, UC_CGM_AC7_DC3_DIV);
#endif
#if (UC_CGM_AC7_DC4_INIT)
	uc_cgm_SetAuxClockDiv(7, 4, UC_CGM_AC7_DC4_DIV);
#endif
#if (UC_CGM_AC7_DC5_INIT)
	uc_cgm_SetAuxClockDiv(7, 5, UC_CGM_AC7_DC5_DIV);
#endif

/* CGM AUX8 */
#if (UC_CGM_AC8_INIT)
	uc_cgm_SelectAuxClock(8, UC_CGM_AC8_CLOCK_INPUT);
#endif
#if (UC_CGM_AC8_DC0_INIT)
	uc_cgm_SetAuxClockDiv(8, 0, UC_CGM_AC8_DC0_DIV);
#endif
#if (UC_CGM_AC8_DC1_INIT)
	uc_cgm_SetAuxClockDiv(8, 1, UC_CGM_AC8_DC1_DIV);
#endif
#if (UC_CGM_AC8_DC2_INIT)
	uc_cgm_SetAuxClockDiv(8, 2, UC_CGM_AC8_DC2_DIV);
#endif
#if (UC_CGM_AC8_DC3_INIT)
	uc_cgm_SetAuxClockDiv(8, 3, UC_CGM_AC8_DC3_DIV);
#endif
#if (UC_CGM_AC8_DC4_INIT)
	uc_cgm_SetAuxClockDiv(8, 4, UC_CGM_AC8_DC4_DIV);
#endif
#if (UC_CGM_AC8_DC5_INIT)
	uc_cgm_SetAuxClockDiv(8, 5, UC_CGM_AC8_DC5_DIV);
#endif

/* CGM AUX9 */
#if (UC_CGM_AC9_INIT)
	uc_cgm_SelectAuxClock(9, UC_CGM_AC9_CLOCK_INPUT);
#endif
#if (UC_CGM_AC9_DC0_INIT)
	uc_cgm_SetAuxClockDiv(9, 0, UC_CGM_AC9_DC0_DIV);
#endif
#if (UC_CGM_AC9_DC1_INIT)
	uc_cgm_SetAuxClockDiv(9, 1, UC_CGM_AC9_DC1_DIV);
#endif
#if (UC_CGM_AC9_DC2_INIT)
	uc_cgm_SetAuxClockDiv(9, 2, UC_CGM_AC9_DC2_DIV);
#endif
#if (UC_CGM_AC9_DC3_INIT)
	uc_cgm_SetAuxClockDiv(9, 3, UC_CGM_AC9_DC3_DIV);
#endif
#if (UC_CGM_AC9_DC4_INIT)
	uc_cgm_SetAuxClockDiv(9, 4, UC_CGM_AC9_DC4_DIV);
#endif
#if (UC_CGM_AC9_DC5_INIT)
	uc_cgm_SetAuxClockDiv(9, 5, UC_CGM_AC9_DC5_DIV);
#endif

/* CGM AUX10 */
#if (UC_CGM_AC10_INIT)
	uc_cgm_SelectAuxClock(10, UC_CGM_AC10_CLOCK_INPUT);
#endif
#if (UC_CGM_AC10_DC0_INIT)
	uc_cgm_SetAuxClockDiv(10, 0, UC_CGM_AC10_DC0_DIV);
#endif
#if (UC_CGM_AC10_DC1_INIT)
	uc_cgm_SetAuxClockDiv(10, 1, UC_CGM_AC10_DC1_DIV);
#endif
#if (UC_CGM_AC10_DC2_INIT)
	uc_cgm_SetAuxClockDiv(10, 2, UC_CGM_AC10_DC2_DIV);
#endif
#if (UC_CGM_AC10_DC3_INIT)
	uc_cgm_SetAuxClockDiv(10, 3, UC_CGM_AC10_DC3_DIV);
#endif
#if (UC_CGM_AC10_DC4_INIT)
	uc_cgm_SetAuxClockDiv(10, 4, UC_CGM_AC10_DC4_DIV);
#endif
#if (UC_CGM_AC10_DC5_INIT)
	uc_cgm_SetAuxClockDiv(10, 5, UC_CGM_AC10_DC5_DIV);
#endif

/* CGM AUX11 */
#if (UC_CGM_AC11_INIT)
	uc_cgm_SelectAuxClock(11, UC_CGM_AC11_CLOCK_INPUT);
#endif
#if (UC_CGM_AC11_DC0_INIT)
	uc_cgm_SetAuxClockDiv(11, 0, UC_CGM_AC11_DC0_DIV);
#endif
#if (UC_CGM_AC11_DC1_INIT)
	uc_cgm_SetAuxClockDiv(11, 1, UC_CGM_AC11_DC1_DIV);
#endif
#if (UC_CGM_AC11_DC2_INIT)
	uc_cgm_SetAuxClockDiv(11, 2, UC_CGM_AC11_DC2_DIV);
#endif
#if (UC_CGM_AC11_DC3_INIT)
	uc_cgm_SetAuxClockDiv(11, 3, UC_CGM_AC11_DC3_DIV);
#endif
#if (UC_CGM_AC11_DC4_INIT)
	uc_cgm_SetAuxClockDiv(11, 4, UC_CGM_AC11_DC4_DIV);
#endif
#if (UC_CGM_AC11_DC5_INIT)
	uc_cgm_SetAuxClockDiv(11, 5, UC_CGM_AC11_DC5_DIV);
#endif

/* CGM AUX12 */
#if (UC_CGM_AC12_INIT)
	uc_cgm_SelectAuxClock(12, UC_CGM_AC12_CLOCK_INPUT);
#endif
#if (UC_CGM_AC12_DC0_INIT)
	uc_cgm_SetAuxClockDiv(12, 0, UC_CGM_AC12_DC0_DIV);
#endif
#if (UC_CGM_AC12_DC1_INIT)
	uc_cgm_SetAuxClockDiv(12, 1, UC_CGM_AC12_DC1_DIV);
#endif
#if (UC_CGM_AC12_DC2_INIT)
	uc_cgm_SetAuxClockDiv(12, 2, UC_CGM_AC12_DC2_DIV);
#endif
#if (UC_CGM_AC12_DC3_INIT)
	uc_cgm_SetAuxClockDiv(12, 3, UC_CGM_AC12_DC3_DIV);
#endif
#if (UC_CGM_AC12_DC4_INIT)
	uc_cgm_SetAuxClockDiv(12, 4, UC_CGM_AC12_DC4_DIV);
#endif
#if (UC_CGM_AC12_DC5_INIT)
	uc_cgm_SetAuxClockDiv(12, 5, UC_CGM_AC12_DC5_DIV);
#endif
}
